from typing import List, Optional
from xpath_helper.filter import ValidExpressionFilter

"""
XPathHelper provides a simple and chainnable API to build complicated XPath queries without the hassle.
After building your XPath query, pass it to the <code>str</code> method  to get the corresponding XPath string.
"""
class XPathHelper:
    
    sb: List[str] = []

    def __init__(self, currentPath: Optional[List[str]]=None):
        """Creates an instance of XPathHelper.

        Args:
            currentPath (list[string]): Current path
        """
        if (currentPath != None):
            self.sb = currentPath

        else:
            self.__append_local_path()

    def empty(self):
        """Empties the current path.
        """
        self.sb = []
        self.__append_local_path()

    def __str__(self) -> str:
        """Returns the corresponding Xpath query.

        Returns:
            str: the string of the corresponding XPath query
        """
        return "".join(self.sb)

    ############## General commands ##############

    def get_parent(self) -> 'XPathHelper' :
        """Select the parent of the current element.

        Returns:
            XPathHelper: a new instance of XPathHelper
        """
        return XPathHelper(self.sb + ["/.."])

    def get_element_by_xpath(self, xpath : str) -> 'XPathHelper' :
        """Select an element with an XPath selector <code>xpath</code>.

        Args:
            xpath (str): [description]

        Returns:
            XPathHelper: a new instance of XPathHelper
        """
        return XPathHelper(self.sb + [xpath])

    ############## Descendant axis ##############
    # The descendant axis retrieves all nodes below the node in reference no matter the depth.
    #############################################

    def get_descendant(self, filter: Optional[ValidExpressionFilter]=None) -> 'XPathHelper' :
        """Select the nodes filtered by <code>filter</code>, below the node in reference no matter the depth.

        Args:
            filter (Filter): filter to apply

        Returns:
            XPathHelper: a new instance of XPathHelper
        """
        return XPathHelper(self.sb + ["//*" + self.__compute_filter(filter)])

    def get_element(self, filter: Optional[ValidExpressionFilter]=None) -> 'XPathHelper' :
        """Select the nodes filtered by <code>filter</code>, below the node in reference no matter the depth.
        This self method is a synonym of <code>get_descendant</code>.

        Args:
            filter (Filter): filter to apply

        Returns:
            XPathHelper: a new instance of XPathHelper
        """
        return self.get_descendant(filter)

    def get_descendant_by_tag(self, tag: str, filter: Optional[ValidExpressionFilter]=None) -> 'XPathHelper' :
        """Select the nodes with tag <code>tag</code> filtered by <code>filter</code>, below the node in reference no matter the depth.

        Args:
            tag (str): tag name
            filter (Filter): filter to apply

        Returns:
            XPathHelper: a new instance of XPathHelper
        """
        return XPathHelper(self.sb + ["//" + tag + self.__compute_filter(filter)])

    def get_element_by_tag(self, tag: str, filter: Optional[ValidExpressionFilter]=None) -> 'XPathHelper' :
        """Select the nodes with tag <code>tag</code> filtered by <code>filter</code>, below the node in reference no matter the depth.
        This method is a synonym of <code>get_descendant_by_tag</code>.

        Args:
            tag (str): tag name
            filter (Filter): filter to apply

        Returns:
            XPathHelper: a new instance of XPathHelper
        """
        return self.get_descendant_by_tag(tag, filter)

    def get_descendant_by_svg_tag(self, svg_tag: str, filter: Optional[ValidExpressionFilter]=None) -> 'XPathHelper':
        """Select the SVG nodes with SVG tag <code>svg_tag</code> filtered by <code>filter</code>, below the node in reference no matter the depth.

        Args:
            svg_tag (str): SVG tag name
            filter (Filter): filter to apply

        Returns:
            XPathHelper: a new instance of XPathHelper
        """
        return XPathHelper(self.sb + ["//*[local-name() = '" + svg_tag + "']" + self.__compute_filter(filter)])

    def get_element_by_svg_tag(self, svg_tag: str, filter: Optional[ValidExpressionFilter]=None) -> 'XPathHelper':
        """Select the SVG nodes with SVG tag <code>svg_tag</code> filtered by <code>filter</code>, below the node in reference no matter the depth.
       This method is a synonym of <code>get_descendant_by_svg_tag</code>.

        Args:
            svg_tag (str): SVG tag name
            filter (Filter): filter to apply

        Returns:
            XPathHelper: a new instance of XPathHelper
        """
        return self.get_descendant_by_svg_tag(svg_tag, filter)

    ############## Descendant-or-self axis ##############
    # The descendant-or-self axis returns all nodes below the current node,
    # but also returns the node in reference.
    ####################################################

    def get_descendant_or_self(self, filter: Optional[ValidExpressionFilter]=None) -> 'XPathHelper' :
        """Select the nodes filtered by <code>filter</code>, below the current node, but also returns the node in reference.

        Args:
            filter (Filter): filter to apply

        Returns:
            XPathHelper: a new instance of XPathHelper
        """
        return XPathHelper(self.sb + ["/descendant-or-self::*" + self.__compute_filter(filter)])

    def get_descendant_or_self_by_tag(self, tag: str, filter: Optional[ValidExpressionFilter]=None) -> 'XPathHelper' :
        """Select the nodes with tag <code>tag</code> filtered by <code>filter</code>, below the current node, but also returns the node in reference.

        Args:
            tag (str): tag name
            filter (Filter): filter to apply

        Returns:
            XPathHelper: a new instance of XPathHelper
        """
        return XPathHelper(self.sb + ["/descendant-or-self::" + tag + "" + self.__compute_filter(filter)])

    def get_descendant_or_self_by_svg_tag(self, svg_tag: str, filter: Optional[ValidExpressionFilter]=None) -> 'XPathHelper':
        """Select the SVG nodes with SVG tag <code>svg_tag</code> filtered by <code>filter</code>, below the current node, but also returns the node in reference.

        Args:
            svg_tag (str): SVG tag name
            filter (Filter): filter to apply

        Returns:
            XPathHelper: a new instance of XPathHelper
        """
        return XPathHelper(self.sb + ["/descendant-or-self::*[local-name() = '" +
                                      svg_tag +
                                      "']" +
                                      self.__compute_filter(filter)])

    ############## Child axis ##############
    # The child axis returns the nodes immediately below the node in reference.
    #######################################

    def get_child(self, filter: Optional[ValidExpressionFilter]=None) -> 'XPathHelper' :
        """Select the nodes filtered by <code>filter</code> immediately below the node in reference.

        Args:
            filter (Filter): filter to apply

        Returns:
            XPathHelper: a new instance of XPathHelper
        """
        return XPathHelper(self.sb + ["/*" + self.__compute_filter(filter)])

    def get_child_by_tag(self, tag: str, filter: Optional[ValidExpressionFilter]=None) -> 'XPathHelper' :
        """Select the nodes with tag <code>tag</code> filtered by <code>filter</code>, immediately below the node in reference.

        Args:
            tag (str): tag name
            filter (Filter): filter to apply

        Returns:
            XPathHelper: a new instance of XPathHelper
        """
        return XPathHelper(self.sb + ["/" + tag + "" + self.__compute_filter(filter)])

    def get_child_by_svg_tag(self, svg_tag: str, filter: Optional[ValidExpressionFilter]=None) -> 'XPathHelper':
        """Select the SVG nodes with SVG tag <code>svg_tag</code> filtered by <code>filter</code>, immediately below the node in reference.

        Args:
            svg_tag (str): SVG tag name
            filter (Filter): filter to apply

        Returns:
            XPathHelper: a new instance of XPathHelper
        """
        return XPathHelper(self.sb + ["/*[local-name() = '" + svg_tag + "']" + self.__compute_filter(filter)])

    ############## Ancestor axis ##############
    # The ancestor axis returns all the nodes that are ancestors,
    # or the parent, and the parent's parent, and so on, to the node in reference.
    ###########################################

    def get_ancestor(self, filter: Optional[ValidExpressionFilter]=None) -> 'XPathHelper' :
        """Select the nodes filtered by <code>filter</code>,
        that are ancestors, or the parent, and the parent's parent, and so on, to the node in reference.

        Args:
            filter (Filter): filter to apply

        Returns:
            XPathHelper: a new instance of XPathHelper
        """
        return XPathHelper(self.sb + ["/ancestor::*" + self.__compute_filter(filter)])

    def get_ancestor_by_tag(self, tag: str, filter: Optional[ValidExpressionFilter]=None) -> 'XPathHelper' :
        """Select the nodes with tag <code>tag</code> filtered by <code>filter</code>,
        that are ancestors, or the parent, and the parent's parent, and so on, to the node in reference.

        Args:
            tag (str): tag name
            filter (Filter): filter to apply

        Returns:
            XPathHelper: a new instance of XPathHelper
        """
        return XPathHelper(self.sb + ["/ancestor::" + tag + "" + self.__compute_filter(filter)])

    def get_ancestor_by_svg_tag(self, svg_tag: str, filter: Optional[ValidExpressionFilter]=None) -> 'XPathHelper':
        """Select the SVG nodes with SVG tag <code>svg_tag</code> filtered by <code>filter</code>,
        that are ancestors, or the parent, and the parent's parent, and so on, to the node in reference.

        Args:
            svg_tag (str): SVG tag name
            filter (Filter): filter to apply

        Returns:
            XPathHelper: a new instance of XPathHelper
        """
        return XPathHelper(self.sb + ["/ancestor::*[local-name() = '" +
                                      svg_tag +
                                      "']" +
                                      self.__compute_filter(filter)])

    ############## Ancestor-or-self axis ##############
    # The ancestor-or-self axis returns all nodes that are ancestors,
    # or the parent, and the parent's parent, and so on, including the node in reference.
    ##############******##############

    def get_ancestor_or_self(self, filter: Optional[ValidExpressionFilter]=None) -> 'XPathHelper' :
        """Select the nodes filtered by <code>filter</code>,
        that are ancestors, or the parent, and the parent's parent, and so on, including the node in reference.

        Args:
            filter (Filter): filter to apply

        Returns:
            XPathHelper: a new instance of XPathHelper
        """
        return XPathHelper(self.sb + ["/ancestor-or-self::*" + self.__compute_filter(filter)])

    def get_ancestor_or_self_by_tag(self, tag: str, filter: Optional[ValidExpressionFilter]=None) -> 'XPathHelper' :
        """Select the nodes with tag <code>tag</code> filtered by <code>filter</code>,
        that are ancestors, or the parent, and the parent's parent, and so on, including the node in reference.

        Args:
            tag (str): tag name
            filter (Filter): filter to apply

        Returns:
            XPathHelper: a new instance of XPathHelper
        """
        return XPathHelper(self.sb + ["/ancestor-or-self::" + tag + "" + self.__compute_filter(filter)])

    def get_ancestor_or_self_by_svg_tag(self, svg_tag: str, filter: Optional[ValidExpressionFilter]=None) -> 'XPathHelper':
        """Select the SVG nodes with SVG tag <code>svg_tag</code> filtered by <code>filter</code>,
        that are ancestors, or the parent, and the parent's parent, and so on, including the node in reference.

        Args:
            svg_tag (str): SVG tag name
            filter (Filter): filter to apply

        Returns:
            XPathHelper: a new instance of XPathHelper
        """
        return XPathHelper(self.sb + ["/ancestor-or-self::*[local-name() = '" +
                                      svg_tag +
                                      "']" +
                                      self.__compute_filter(filter)])

    ############## Following axis ##############
    # The following axis selects all nodes no matter the depth, that are located on parent-level
    # and also after (following) its parent of the node in reference.
    ############################################

    def get_following(self, filter: Optional[ValidExpressionFilter]=None) -> 'XPathHelper' :
        """Select the nodes filtered by <code>filter</code>, that are located on parent-level
        and also after (following) its parent of the node in reference.

        Args:
            filter (Filter): filter to apply

        Returns:
            XPathHelper: a new instance of XPathHelper
        """
        return XPathHelper(self.sb + ["/following::*" + self.__compute_filter(filter)])

    def get_following_by_tag(self, tag: str, filter: Optional[ValidExpressionFilter]=None) -> 'XPathHelper' :
        """Select the nodes with tag <code>tag</code> filtered by <code>filter</code>,
        that are located on parent-level and also after (following) its parent of the node in reference.

        Args:
            tag (str): tag name
            filter (Filter): filter to apply

        Returns:
            XPathHelper: a new instance of XPathHelper
        """
        return XPathHelper(self.sb + ["/following::" + tag + "" + self.__compute_filter(filter)])

    def get_following_by_svg_tag(self, svg_tag: str, filter: Optional[ValidExpressionFilter]=None) -> 'XPathHelper':
        """Select the SVG nodes with SVG tag <code>svg_tag</code> filtered by <code>filter</code>,
        that are located on parent-level and also after (following) its parent of the node in reference.

        Args:
            svg_tag (str): SVG tag name
            filter (Filter): filter to apply

        Returns:
            XPathHelper: a new instance of XPathHelper
        """
        return XPathHelper(self.sb + ["/following::*[local-name() = '" +
                                      svg_tag +
                                      "']" +
                                      self.__compute_filter(filter)])

    ############## Following-sibling axis ##############
    # The following-sibling axis selects all nodes that are located on the same level who are located after (following) the node in reference.
    ####################################################

    def get_following_sibling(self, filter: Optional[ValidExpressionFilter]=None) -> 'XPathHelper' :
        """Select the nodes filtered by <code>filter</code>,
        that are located on the same level who are located after (following) the node in reference.

        Args:
            filter (Filter): filter to apply

        Returns:
            XPathHelper: a new instance of XPathHelper
        """
        return XPathHelper(self.sb + ["/following-sibling::*" + self.__compute_filter(filter)])

    def get_following_sibling_by_tag(self, tag: str, filter: Optional[ValidExpressionFilter]=None) -> 'XPathHelper' :
        """Select the nodes with tag <code>tag</code> filtered by <code>filter</code>,
        that are located on the same level who are located after (following) the node in reference.

        Args:
            tag (str): tag name
            filter (Filter): filter to apply

        Returns:
            XPathHelper: a new instance of XPathHelper
        """
        return XPathHelper(self.sb + ["/following-sibling::" + tag + "" + self.__compute_filter(filter)])

    def get_following_sibling_by_svg_tag(self, svg_tag: str, filter: Optional[ValidExpressionFilter]=None) -> 'XPathHelper':
        """Select the SVG nodes with SVG tag <code>svg_tag</code> filtered by <code>filter</code>,
        that are located on the same level who are located after (following) the node in reference.

        Args:
            svg_tag (str): SVG tag name
            filter (Filter): filter to apply

        Returns:
            XPathHelper: a new instance of XPathHelper
        """
        return XPathHelper(self.sb + ["/following-sibling::*[local-name() = '" +
                                      svg_tag +
                                      "']" +
                                      self.__compute_filter(filter)])

    ############## Preceding axis ##############
    # The preceding axis selects all nodes no matter the depth,
    # that are located on parent-level and also before (preceding) its parent of the node in reference.
    ############################################

    def get_preceding(self, filter: Optional[ValidExpressionFilter]=None) -> 'XPathHelper' :
        """Select the nodes filtered by <code>filter</code>, that are located on parent-level
        and also before (preceding) its parent of the node in reference.

        Args:
            filter (Filter): filter to apply

        Returns:
            XPathHelper: a new instance of XPathHelper
        """
        return XPathHelper(self.sb + ["/preceding::*" + self.__compute_filter(filter)])

    def get_preceding_by_tag(self, tag: str, filter: Optional[ValidExpressionFilter]=None) -> 'XPathHelper' :
        """Select the nodes with tag <code>tag</code> filtered by <code>filter</code>,
        that are located on parent-level and also before (preceding) its parent of the node in reference.

        Args:
            tag (str): tag name
            filter (Filter): filter to apply

        Returns:
            XPathHelper: a new instance of XPathHelper
        """
        return XPathHelper(self.sb + ["/preceding::" + tag + "" + self.__compute_filter(filter)])

    def get_preceding_by_svg_tag(self, svg_tag: str, filter: Optional[ValidExpressionFilter]=None) -> 'XPathHelper':
        """Select the SVG nodes with SVG tag <code>svg_tag</code> filtered by <code>filter</code>,
        that are located on parent-level and also before (preceding) its parent of the node in reference.

        Args:
            svg_tag (str): SVG tag name
            filter (Filter): filter to apply

        Returns:
            XPathHelper: a new instance of XPathHelper
        """
        return XPathHelper(self.sb + ["/preceding::*[local-name() = '" +
                                      svg_tag +
                                      "']" +
                                      self.__compute_filter(filter)])

    ############## Preceding-sibling axis ##############
    # The preceding axis selects all nodes that are located on the same level
    # who are also located before (preceding) the node in reference.
    ##############*******##############

    def get_preceding_sibling(self, filter: Optional[ValidExpressionFilter]=None) -> 'XPathHelper' :
        """Select the nodes filtered by <code>filter</code>, that are located on the same level
        who are also located before (preceding) the node in reference.

        Args:
            filter (Filter): filter to apply

        Returns:
            XPathHelper: a new instance of XPathHelper
        """
        return XPathHelper(self.sb + ["/preceding-sibling::*" + self.__compute_filter(filter)])

    def get_preceding_sibling_by_tag(self, tag: str, filter: Optional[ValidExpressionFilter]=None) -> 'XPathHelper' :
        """Select the nodes with tag <code>tag</code> filtered by <code>filter</code>,
        that are located on the same level who are also located before (preceding) the node in reference.

        Args:
            tag (string): [description]
            filter (Filter): filter to apply

        Returns:
            XPathHelper: a new instance of XPathHelper
        """
        return XPathHelper(self.sb + ["/preceding-sibling::" + tag + "" + self.__compute_filter(filter)])

    def get_preceding_sibling_by_svg_tag(self, svg_tag: str, filter: Optional[ValidExpressionFilter]=None) -> 'XPathHelper':
        """Select the SVG nodes with SVG tag <code>svg_tag</code> filtered by <code>filter</code>,
        that are located on the same level who are also located before (preceding) the node in reference.

        Args:
            svg_tag (str): SVG tag name
            filter (Filter): filter to apply

        Returns:
            XPathHelper: a new instance of XPathHelper
        """
        return XPathHelper(self.sb + ["/preceding-sibling::*[local-name() = '" +
                                      svg_tag +
                                      "']" +
                                      self.__compute_filter(filter)])

    def __append_local_path(self):
        """Adds the local path.

        Returns:
            XPathHelper: an instance of XPathHelper with the local path appened.
        """
        return XPathHelper(self.sb + ["."])

    def __compute_filter(self, filter: Optional[ValidExpressionFilter]=None) -> 'XPathHelper' :
        """Add the given filter to the current xpath expression.

        Args:
            filter (Filter): filter to apply

        Returns:
            str: filter as an XPath string
        """
        suffix = ""
        if (filter != None and not filter.is_empty()):
            expression = str(filter)
            suffix = "[" + expression + "]"

        return suffix


import functools
from typing import List, Optional

"""
The following Filter classes provide a simple, chainable and decomposable api
to create XPath filter expression
"""

"""
Shortcut to design any attribute name
"""
ANY_ATTRIBUTE = "*"


"""
XPath Filter containing a valid expression.
"""


class ValidExpressionFilter:
    sb: List[str] = []

    def __init__(self, current_path: Optional[List[str]]=None):
        """Creates an instance of ValidExpressionFilter.

        Args:
            currentPath (list[string]): Current filter path
        """
        if (current_path != None):
            self.sb = current_path

    def and_operator(self, *filters: 'EmptyFilter') -> 'ValidExpressionFilter':
        """Adds one or more filter expression to the current one with the AND logical operator.

        Args:
            filters (list[Filter]): List of filters List of filters that will be added with the AND logical operator

        Returns:
            ValidExpressionFilter: a new instance of ValidExpressionFilter with the newly formed expression.
        """
        expression = ""
        if (len(self.sb) != 0):
            expression += " and "

        expression += "("
        expression += functools.reduce(lambda acc, filter: acc + add_openrand(
            filter, " and ", filter == filters[-1]), filters, "")
        expression += ")"
        return ValidExpressionFilter(self.sb + [expression])

    def or_operator(self, *filters: 'EmptyFilter') -> 'ValidExpressionFilter':
        """Adds one or more filter expression to the current one with the OR logical operator.

        Args:
            filters (list[Filter]): List of filters that will be added with the OR logical operator

        Returns:
            ValidExpressionFilter: a new instance of ValidExpressionFilter with the newly formed expression.
        """
        expression = ""
        if (len(self.sb) != 0):
            expression += " or "

        expression += "("
        expression += functools.reduce(lambda acc, filter: acc + add_openrand(
            filter, " or ", filter == filters[-1]), filters, "")
        expression += ")"
        return ValidExpressionFilter(self.sb + [expression])

    def __str__(self):
        """Returns the Filter as a valid XPath filter expression.

        Returns:
            str: the string of the corresponding XPath filter
        """
        return "".join(self.sb)

    def empty(self):
        """Empties the current path.
        """
        self.sb = []

    def is_empty(self) -> bool:
        """Returns true if filter is empty.

        Returns:
            bool: true if filter is empty
        """
        return len(self.sb) == 0


"""
Empty XPath filter.
"""


class EmptyFilter(ValidExpressionFilter):
    def __init__(self, currentPath=None):
        """Creates an instance of Filter.

        Args:
            currentPath (list[string]): Current filter path
        """
        super().__init__(currentPath)

    def has_attribute(self, attribute: str) -> ValidExpressionFilter:
        """Selects the nodes that have the attribute <code>attribute</code>.

        Args:
            attribute (str): attribute name

        Returns:
            ValidExpressionFilter: a new instance of ValidExpressionFilter with the newly formed expression.
        """
        return ValidExpressionFilter(self.sb + ["@" + attribute])

    def attribute_contains(self, attribute: str, value: str) -> ValidExpressionFilter:
        """Selects the nodes with the attribute <code>attribute</code> containing the value <code><value</code>.

        Args:
            attribute (str): attribute name
            value (str): attribute value

        Returns:
            ValidExpressionFilter: a new instance of ValidExpressionFilter with the newly formed expression.
        """
        return ValidExpressionFilter(self.sb + ["contains(@" + attribute + ", " + replace_apostrophes(value) + ")",
                                                ])

    def attribute_equals(self, attribute: str, value: str) -> ValidExpressionFilter:
        """Selects the nodes with the attribute <code>attribute</code>, whose value equals <code><value</code>.

        Args:
            attribute (str): attribute name
            value (str | int | float): attribute value

        Returns:
            ValidExpressionFilter: a new instance of ValidExpressionFilter with the newly formed expression.
        """
        return ValidExpressionFilter(self.sb + ["@" + attribute + "=" + replace_apostrophes(value) + "",
                                                ])

    def attribute_not_equals(self, attribute: str, value: str) -> ValidExpressionFilter:
        """Selects the nodes with the attribute <code>attribute</code>, whose value doesn't equal <code><value</code>.

        Args:
            attribute (str): attribute name
            value (str | int | float): attribute value

        Returns:
            ValidExpressionFilter: a new instance of ValidExpressionFilter with the newly formed expression.
        """
        return ValidExpressionFilter(self.sb + ["@" + attribute + "!=" + replace_apostrophes(value)])

    def attribute_less_than(self, attribute: str, value: str) -> ValidExpressionFilter:
        """Selects the nodes with the attribute <code>attribute</code>, whose value is less than <code><value<code>.

        Args:
            attribute (str): attribute name
            value (int | float): attribute value

        Returns:
            ValidExpressionFilter: a new instance of ValidExpressionFilter with the newly formed expression.
        """
        return ValidExpressionFilter(self.sb + ["@" + attribute + "<" + str(value)])

    def attribute_less_than_or_equal_to(self, attribute: str, value: str) -> ValidExpressionFilter:
        """Selects the nodes with the attribute <code>attribute</code>, whose value is less than or equal to <code><value</code>.

        Args:
            attribute (str): attribute name
            value (int | float): attribute value

        Returns:
            ValidExpressionFilter: a new instance of ValidExpressionFilter with the newly formed expression.
        """
        return ValidExpressionFilter(self.sb + ["@" + attribute + "<=" + str(value)])

    def attribute_greater_than(self, attribute: str, value: str) -> ValidExpressionFilter:
        """Selects the nodes with the attribute <code>attribute</code>, whose value is greater than <code><value</code>.

        Args:
            attribute (str): attribute name
            value (int | float): attribute value

        Returns:
            ValidExpressionFilter: a new instance of ValidExpressionFilter with the newly formed expression.
        """
        return ValidExpressionFilter(self.sb + ["@" + attribute + ">" + str(value)])

    def attribute_greater_than_or_equal_to(self, attribute: str, value: str) -> ValidExpressionFilter:
        """Selects the nodes with the attribute <code>attribute</code>, whose value is greater than or equal to <code><value</code>.

        Args:
            attribute (str): attribute name
            value (int | float): attribute value

        Returns:
            ValidExpressionFilter: a new instance of ValidExpressionFilter with the newly formed expression.
        """
        return ValidExpressionFilter(self.sb + ["@" + attribute + ">=" + str(value)])

    def value_contains(self, value: str) -> ValidExpressionFilter:
        """Selects the nodes containing the value <code><value</code>.

        Args:
            value (str): value

        Returns:
            ValidExpressionFilter: a new instance of ValidExpressionFilter with the newly formed expression.
        """
        return ValidExpressionFilter(self.sb + ["text()[contains(., " + replace_apostrophes(value) + ")]"])

    def value_equals(self, value: str) -> ValidExpressionFilter:
        """Selects the nodes whose value equals <code><value</code>.

        Args:
            value (str | int | float): value

        Returns:
            ValidExpressionFilter: a new instance of ValidExpressionFilter with the newly formed expression.
        """
        return ValidExpressionFilter(self.sb + ["text() = " + replace_apostrophes(value)])

    def value_not_equals(self, value: str) -> ValidExpressionFilter:
        """Selects the nodes with whose value doesn't equal <code><value</code>.

        Args:
            value (str | int | float): value

        Returns:
            ValidExpressionFilter: a new instance of ValidExpressionFilter with the newly formed expression.
        """
        return ValidExpressionFilter(self.sb + ["text() !=" + replace_apostrophes(value)])

    def value_less_than(self, value: str) -> ValidExpressionFilter:
        """Selects the nodes whose value is less than <code><value</code>.

        Args:
            value (int | float): value

        Returns:
            ValidExpressionFilter: a new instance of ValidExpressionFilter with the newly formed expression.
        """
        return ValidExpressionFilter(self.sb + ["text() <" + str(value)])

    def value_less_than_or_equal_to(self, value: str) -> ValidExpressionFilter:
        """Selects the nodes whose value is less than or equal to <code><value</code>.

        Args:
            value (int | float): value

        Returns:
            ValidExpressionFilter: a new instance of ValidExpressionFilter with the newly formed expression.
        """
        return ValidExpressionFilter(self.sb + ["text() <=" + str(value)])

    def value_greater_than(self, value: str) -> ValidExpressionFilter:
        """Selects the nodes  whose value is greater than <code><value</code>.

        Args:
            value (int | float): value

        Returns:
            ValidExpressionFilter: a new instance of ValidExpressionFilter with the newly formed expression.
        """
        return ValidExpressionFilter(self.sb + ["text() >" + str(value)])

    def value_greater_than_or_equal_to(self, value: str) -> ValidExpressionFilter:
        """Selects the nodes whose value is greater than or equal to <code><value</code>.

        Args:
            value (int | float): value

        Returns:
            ValidExpressionFilter: a new instance of ValidExpressionFilter with the newly formed expression.
        """
        return ValidExpressionFilter(self.sb + ["text() >=" + str(value)])

    def get(self, index: int):
        """Selects the node element who is positioned at the <code>index</code> position in its parent children list.

        Args:
            index (int): index of the element in the list

        Returns:
            ValidExpressionFilter: a new instance of ValidExpressionFilter with the newly formed expression.
        """
        return ValidExpressionFilter(self.sb + [str(index)])

    def get_first(self):
        """Selects the node element who is positioned first in its parent children list.

        Returns:
            ValidExpressionFilter: a new instance of ValidExpressionFilter with the newly formed expression.
        """
        return ValidExpressionFilter(self.sb + ["1"])

    def get_last(self):
        """Selects the node element who is positioned last in its parent children list.

        Returns:
            ValidExpressionFilter: a new instance of ValidExpressionFilter with the newly formed expression.
        """
        return ValidExpressionFilter(self.sb + ["last()"])

    def not_operator(self, filter: ValidExpressionFilter):
        """Reverses the filter <code>filter</code>. Returns true when the filter returns false and true when the filter returns false.

        Args:
            filter ([type]): [description]

        Returns:
            ValidExpressionFilter: a new instance of ValidExpressionFilter with the newly formed expression.
        """
        return ValidExpressionFilter(self.sb + ["not( " + add_openrand(filter) + " )"])


def add_openrand(filter: EmptyFilter, separator="", is_last: Optional[bool] = True) -> str:
    """Adds operand between filters

    Args:
        filter (Filter): a filter
        separator (str, optional): " and " of " or " separator. Defaults to "".
        isLast (bool, optional): True if it is the last operand. Defaults to True.

    Returns:
        str: opperand with and / or separator
    """
    suffix: str = ""
    if filter and not filter.is_empty():
        expression = str(filter)
        suffix = expression
        if not is_last:
            suffix += separator
    return suffix


def replace_apostrophes(input: str) -> str:
    """Treats the presence of apostrophes so it doesn't break the XPath filter expression.

    Args:
        input (str | int): input

    Returns:
        str: XPath filter expression with apostrophes handled.
    """
    if not isinstance(input, str):
        return str(input)

    if "'" in input:
        prefix: str = ""
        elements = input.split("'")
        output = "concat("
        for s in elements:
            output += prefix + "'" + s + "'"
            prefix = ',"\'",'

        output += ")"
        return output

    else:
        return "'" + input + "'"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['xpath_helper']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'xpath-helper',
    'version': '0.1.1',
    'description': 'A simple and chainnable API to build complicated XPath queries without the hassle.',
    'long_description': '<p align="center">\n  <img src="https://raw.githubusercontent.com/jrebecchi/xpath-helper/main/docs/_static/logo-with-text.png" height="100px" alt="xpath-helper"/>\n</p>\n<p align="center">\n  <i>Build complicated XPath queries without the hassle - JavaScript & Python</i><br/><br/>\n  <a href="https://jrebecchi.github.io/xpath-helper/python/index.html">\n    <img src="https://img.shields.io/badge/docs-master-blue.svg?style=flat">\n  </a>\n  <a href="https://codecov.io/gh/jrebecchi/xpath-helper">\n    <img src="https://codecov.io/gh/jrebecchi/xpath-helper/branch/main/graph/badge.svg?token=0TJOZ64X10"/>\n  </a>\n  <a href="https://github.com/jrebecchi/xpath-helper/actions/workflows/python-ci.yml">\n    <img src="https://github.com/jrebecchi/xpath-helper/actions/workflows/python-ci.yml/badge.svg">\n  </a>\n</p>\n\nA chainable API to build complex XPath queries along the different [XPath axes](https://jrebecchi.github.io/xpath-helper/xpath-axes.html). Available both in Python and JavaScript.\n\n- [**Documentation**](https://jrebecchi.github.io/xpath-helper/python/index.html) — Consult the quick start guide and the online documentation.\n\n## Installation\n`xpath-helper` requires python 3.5+ and can be installed using pip:\n```bash\npip install xpath-helper\n```\n## Quick-start\nYou can chain method call on the different [XPath axes](https://jrebecchi.github.io/xpath-helper/xpath-axes.html) and easily add filters.\n```python\nfrom xpath_helper import xh, filter\n\n# Find a paragraph <p> containing a CSS class \'very-nice-p\'\np = xh.get_element_by_tag(\'p\', filter.attribute_contains(\'class\', \'very-nice-p\'))\nstr(p) # "//p[contains(@class, \'very-nice-p\')]"\n\n# Find the paragraph that is following the above one\nnext_p = p.get_following_sibling_by_tag(\'p\')\nstr(next_p) # "//p[contains(@class, \'very-nice-p\')]/following-sibling::p"\n\n# Find the modal containing a button with text "Register" \nmodal = xh.get_element(filter.value_equals(\'Register\')).get_ancestor(filter.attribute_equals(\'class\', \'modal\'))\nstr(modal) # "//*[text() = \'Register\']/ancestor::*[@class=\'modal\']"\n\n# An elaborated filter with a boolean expression\nli = xh.get_element_by_tag("li", filter.and_operator(\n  filter.or_operator(\n    filter.value_contains("JavaScript"), filter.value_contains("Python")\n  ),\n  filter.has_attribute("data-description")\n))\nstr(li) # "//li[((text()[contains(., \'JavaScript\')] or text()[contains(., \'Python\')]) and @data-description)]"\n```\n\n## Chaining\n\nXPath natively lets your build complex queries chaining them along its different axes. Read this [article to understand the different XPath axes](https://jrebecchi.github.io/xpath-helper/xpath-axes.html).\n\nThis library let you do exactly the same by chaining method calls along the different axes: [`descendant`](https://jrebecchi.github.io/xpath-helper/python/api.html#xpath_helper.XPathHelper.get_descendant) aliased as [`element`](https://jrebecchi.github.io/xpath-helper/python/api.html#xpath_helper.XPathHelper.get_element), [`descendant-or-self`](https://jrebecchi.github.io/xpath-helper/python/api.html#xpath_helper.XPathHelper.get_descendant_or_self), [`child`](https://jrebecchi.github.io/xpath-helper/python/api.html#xpath_helper.XPathHelper.get_child), [`parent`](https://jrebecchi.github.io/xpath-helper/python/api.html#xpath_helper.XPathHelper.get_parent), [`ancestor`](https://jrebecchi.github.io/xpath-helper/python/api.html#xpath_helper.XPathHelper.get_ancestor), [`ancestor-or-self`](https://jrebecchi.github.io/xpath-helper/python/api.html#xpath_helper.XPathHelper.get_ancestor_or_self), [`preceding`](https://jrebecchi.github.io/xpath-helper/python/api.html#xpath_helper.XPathHelper.get_preceding), [`preceding-sibling`](https://jrebecchi.github.io/xpath-helper/python/api.html#xpath_helper.XPathHelper.get_preceding_sibling), [`following`](https://jrebecchi.github.io/xpath-helper/python/api.html#xpath_helper.XPathHelper.get_following), [`following-sibling`](https://jrebecchi.github.io/xpath-helper/python/api.html#xpath_helper.XPathHelper.get_following_sibling).\n\nFor each axis, xpath-helper provides 3 methods, like for instance `get_element(filter)`, `get_element_by_tag(tag, filter)`, `get_element_by_svg_tag(svgTag, filter)` for the `descendant`  axis aliased as `element`.\n\n***The complete filter API can be found [here](https://jrebecchi.github.io/xpath-helper/python/api.html#module-xpath_helper).***\n\n\n```python\nfrom xpath_helper import xh, filter\n\n# Find an element into the page, move to its parent, \n# find a brother node of the parent positioned after it.\nel = xh.get_element_by_tag(\n    \'p\', filter.attribute_contains(\'class\', \'very-nice-p\')\n).get_parent().get_following_sibling_by_tag(\'p\')\nstr(el) # "//p[contains(@class, \'very-nice-p\')]/../following-sibling::p"\n\n# Find an element into the page, move to its ancestor \n# containing \'very-nice-p\' ass CSS class, \n# find a brother node of the ancestor positioned before it.\nel = xh.get_element_by_tag(\n    \'p\', filter.attribute_contains(\'class\', \'very-nice-p\')\n).get_ancestor_by_tag(\n  \'div\'\n).get_preceding_sibling(filter.has_attribute(\'data-foo-bar\'))\nstr(el) # "//p[contains(@class, \'very-nice-p\')]/../following-sibling::p//p[contains(@class, \'very-nice-p\')]/ancestor::div/preceding-sibling::*[@data-foo-bar]"\n```\n\nIt is also possible to keep a relative path in a variable and re-use it after.\n```python\nfrom xpath_helper import xh, filter\n# Store the path of a modal window\nmodal = xh.get_element(filter.attribute_contains(\'class\', \'modal\'))\n# Find the Submit button inside the modal window\nsubmit_button = modal.get_element_by_tag(\'button\', filter.value_equals(\'Submit\'))\n# Find the Cancel button inside the modal window\ncancel_button = modal.get_element_by_tag(\'button\', filter.value_equals(\'Cancel\'))\n```\n## Filters\nTo select elements more precisely you can add filters: [on attributes](https://jrebecchi.github.io/xpath-helper/python/api.html#xpath_helper.filter.EmptyFilter.attribute_contains), on [element values](https://jrebecchi.github.io/xpath-helper/python/api.html#xpath_helper.filter.EmptyFilter.value_contains), [element position](https://jrebecchi.github.io/xpath-helper/python/api.html#xpath_helper.filter.EmptyFilter.get), and combining them with conditional operators: [`and_operator(...)`](https://jrebecchi.github.io/xpath-helper/python/api.html#xpath_helper.filter.ValidExpressionFilter.and_operator), [`or_operator(...)`](https://jrebecchi.github.io/xpath-helper/python/api.html#xpath_helper.filter.ValidExpressionFilter.or_operator), and [`not_operator(...)`](https://jrebecchi.github.io/xpath-helper/python/api.html#xpath_helper.filter.EmptyFilter.not_operator).\n\n***The complete filter API can be found [here](https://jrebecchi.github.io/xpath-helper/python/api.html#filter-filter).***\n\n### Attributes\nFind below a few examples of filters on attributes.\n\n```python\nfrom xpath_helper import xh, filter\n# Looks for an element that has a class attribute equals to \'foo\'\nel = xh.get_element(filter.attribute_equals(\'class\', \'foo\'))\n# Looks for an element that has a class attribute containing \'bar\'\nel = xh.get_element(filter.attribute_contains(\'class\', \'bar\'))\n# Looks for an element that has the attribute \'alt\'\nimg = xh.get_element_by_tag(\'img\', filter.has_attribute(\'alt\'))\n# Looks for all the li element with a data-attribute superior to 3\nli = xh.get_element_by_tag(\'li\', filter.attribute_greater_than(\'data-index\', 3))\n```\n### Values\nFind below a few examples of filters on node values.\n```python\nfrom xpath_helper import xh, filter\n\n# Looks for a button whose text is \'Submit\'\nbutton = xh.get_element_by_tag(\'button\', filter.value_equals(\'Submit\'))\n# Looks for an element whose text contains \'foobar\'\nel = xh.get_element(filter.value_contains(\'foobar\'))\n# Looks for all the li element with a value superior to 3\nli = xh.get_element_by_tag(\'li\', filter.value_greater_than(3))\n```\n### Position\nFind below a few examples of filters on node position.\n```python\nfrom xpath_helper import xh, filter\n\n# Looks for the first li element in ul list\nfirst = xh.get_element_by_tag(\'ul\').get_element_by_tag(\'li\', filter.get_first())\n# Looks for the first li element in ul list\nlast = xh.get_element_by_tag(\'ul\').get_element_by_tag(\'li\', filter.get_last())\n# Looks for the third li element in ul list\nthird = xh.get_element_by_tag(\'ul\').get_element_by_tag(\'li\', filter.get(3))\n```\n### Conditional expression\nFind below a few examples of filters with conditional expression.\n\n```python\nfrom xpath_helper import xh, filter\n\n# Find an element that has a CSS class \'a-link\' and contains an attribute href\nel = xh.get_element(\n  filter.attribute_contains(\'class\', \'a-link\').and_operator(\n    filter.has_attribute(\'href\')\n  )\n)\nstr(el) # "//*[contains(@class, \'a-link\') and (@href)]"\n\n# Find an element that has a CSS class \'foo\' or a CSS class \'bar\'\nel = xh.get_element(\n  filter.attribute_contains(\'class\', \'foo\').or_operator(\n    filter.attribute_contains(\'class\', \'bar\')\n  )\n)\nstr(el) # "//*[contains(@class, \'foo\') or (contains(@class, \'bar\'))]"\n\n# Build complex logical expression combining and & or\nel = xh.get_element(\n  filter.and_operator(\n    filter.or_operator(\n      filter.value_contains("JavaScript"),\n      filter.value_contains("Pyhton")\n    ),\n      filter.value_contains("package")\n  )\n)\nstr(el) # "//*[((text()[contains(., \'JavaScript\')] or text()[contains(., \'Pyhton\')]) and text()[contains(., \'package\')])]"\n```\n## SVG\nNavigating into SVG elements from an HTML file can be tricky with XPath, that is why a subset of functions have been added. They are all ending with `...by_svg_tag` and can be used as below.\n\n```python\nfrom xpath_helper import xh, filter\n\n# Store the path of a modal window\npath = xh.get_element_by_svg_tag(\'path\', filter.attribute_equals(\'d\', \'M 310 130 L 90 130 L 90 183.63\'))\nstr(path) # "//*[local-name() = \'path\'][@d=\'M 310 130 L 90 130 L 90 183.63\']"\n\n# Find the Submit button inside the modal window\ng = xh.get_element_by_svg_tag(\n  \'path\', filter.attribute_equals(\'id\', \'id-path\')\n).get_ancestor_by_svg_tag(\'g\')\nstr(g) # "//*[local-name() = \'path\'][@id=\'id-path\']/ancestor::*[local-name() = \'g\']"\n```',
    'author': 'Johann Rebecchi',
    'author_email': 'jrebecchi@protonmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/jrebecchi/xpath-helper',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.5',
}


setup(**setup_kwargs)



__version__ = '0.1.1'
__all__ = ['xh', 'filter', 'XPathHelper', 'EmptyFilter']

from xpath_helper.filter import EmptyFilter
from xpath_helper.xpath_helper import XPathHelper
filter = EmptyFilter()
xh = XPathHelper()

